@@ -12,6 +12,7 @@ import re
 
 from tdmclient import ClientAsync, ArrayCache
 from tdmclient.atranspiler import ATranspiler
+from tdmclient.module_thymio import ModuleThymio
 
 
 class TDMConsole(code.InteractiveConsole):
@@ -30,6 +31,8 @@ class TDMConsole(code.InteractiveConsole):
         self.client = None
         self.node = None
 
+        self.event_data_dict = {}
+
         def onevent(fun):
             """Function decorator @onevent for event handlers. The event name
             is given by the function name.
@@ -100,7 +103,8 @@ class TDMConsole(code.InteractiveConsole):
 
             if language == "aseba":
                 # transpile from Python to Aseba
-                src = ATranspiler.simple_transpile(src)
+                transpiler = self.transpile(src, True)
+                src = transpiler.get_output()
             elif language != "python":
                 raise Exception(f"Unsupported language {language}")
 
@@ -112,17 +116,30 @@ class TDMConsole(code.InteractiveConsole):
             self.robot_var_set.clear()
             self.onevent_functions.clear()
 
-        def run():
-            """Run program obtained by robot_code on the robot.
+        def run(wait=None):
+            """Run program obtained by robot_code on the robot. By default, wait
+            to process events until "_exit" is received (call to "exit()" in the
+            robot's program), or return immediately if the program doesn't send
+            any event.
             """
-            src_a = robot_code(language="aseba")
+            src_p = robot_code()
             # compile, load, run, and set scratchpad without checking the result
-            self.run_program(src_a)
+            self.run_program(src_p, language="python", wait=wait)
 
         def stop():
             """Stop the program running on the robot.
             """
-            self.stop_program()
+            self.stop_program(discard_output=True)
+
+        def clear_event_data(event_name=None):
+            """Clear all or named event data.
+            """
+            self.clear_event_data(event_name)
+
+        def get_event_data(event_name=None):
+            """Get list of event data received until now.
+            """
+            return self.get_event_data(event_name)
 
         self.functions = {
             "onevent": onevent,
@@ -131,6 +148,8 @@ class TDMConsole(code.InteractiveConsole):
             "robot_code_new": robot_code_new,
             "run": run,
             "stop": stop,
+            "clear_event_data": clear_event_data,
+            "get_event_data": get_event_data,
         }
 
         if local_var is None:
@@ -156,6 +175,9 @@ class TDMConsole(code.InteractiveConsole):
         self.var_got = set()
         self.var_set = set()
 
+        # enable output upon receiving events
+        self.output_enabled = True
+
     async def init(self, client, node):
         self.client = client
         self.node = node
@@ -174,25 +196,105 @@ class TDMConsole(code.InteractiveConsole):
             sync_var.add(name_py)
         self.sync_var = sync_var
 
-    def run_program(self, src, language="aseba"):
+    @staticmethod
+    def transpile(src, import_thymio=True):
+        """Transpile Python source code to Aseba and returns transpiler.
+
+        Argument:
+            src -- Python source code
+            import_thymio -- if True (default), predefine all Thymio symbols
+        """
+        modules = {
+            "thymio": ModuleThymio()
+        }
+        transpiler = ATranspiler()
+        transpiler.modules = {**transpiler.modules, **modules}
+        if import_thymio:
+            transpiler.set_preamble("""from thymio import *
+""")
+        transpiler.set_source(src)
+        transpiler.transpile()
+        return transpiler
+
+    def clear_event_data(self, event_name=None):
+        if event_name is None:
+            self.event_data_dict = {}
+        elif event_name in self.event_data_dict:
+            del self.event_data_dict[event_name]
+
+    def get_event_data(self, event_name=None):
+        if event_name is None:
+            return self.event_data_dict
+        else:
+            return self.event_data_dict[event_name] if event_name in self.event_data_dict else []
+
+    def run_program(self, src, language="aseba", wait=False, import_thymio=True):
+        print_statements = []
+        events = []
         if language == "python":
             # transpile from Python to Aseba
-            src = ATranspiler.simple_transpile(src)
+            transpiler = self.transpile(src, import_thymio)
+            src = transpiler.get_output()
+            print_statements = transpiler.print_format_strings
+            if len(print_statements) > 0:
+                events.append(("_print", 1 + transpiler.print_max_num_args))
+            if transpiler.has_exit_event:
+                events.append(("_exit", 1))
+            for event_name in transpiler.events:
+                events.append((event_name, transpiler.events[event_name]))
+            if len(events) > 0:
+                ClientAsync.aw(self.node.register_events(events))
         elif language != "aseba":
             raise Exception(f"Unsupported language {language}")
         # compile, load, run, and set scratchpad without checking the result
         error = ClientAsync.aw(self.node.compile(src))
         if error is not None:
             raise Exception(error["error_msg"])
+        exit_received = None  # or exit code once received
+        if wait is None:
+            # wait if there are events to receive
+            wait = len(events) > 0
+        if len(events) > 0 and wait:
+            def on_event_received(node, event_name, event_data):
+                if self.output_enabled:
+                    if event_name == "_exit":
+                        nonlocal exit_received
+                        exit_received = event_data[0]
+                    elif event_name == "_print":
+                        print_id = event_data[0]
+                        print_format, print_num_args = print_statements[print_id]
+                        print_args = tuple(event_data[1 : 1 + print_num_args])
+                        print_str = print_format % print_args
+                        print(print_str)
+                    else:
+                        if len(event_data) > 0:
+                            if event_name not in self.event_data_dict:
+                                self.event_data_dict[event_name] = []
+                            self.event_data_dict[event_name].append(event_data)
+            self.client.clear_event_received_listeners()
+            self.client.add_event_received_listener(on_event_received)
+            ClientAsync.aw(self.node.watch(events=True))
         error = ClientAsync.aw(self.node.run())
         if error is not None:
             raise Exception(f"Error {error['error_code']}")
         self.node.send_set_scratchpad(src)
+        if wait:
+            def wake():
+                return exit_received is not None
+            ClientAsync.aw(self.client.sleep(wake=wake))
+            self.stop_program(discard_output=True)
+            if exit_received:
+                print(f"Exit, status={exit_received}")
 
-    def stop_program(self):
-        error = ClientAsync.aw(self.node.stop())
-        if error is not None:
-            raise Exception(f"Error {error['error_code']}")
+    def stop_program(self, discard_output=False):
+        output_enabled_orig = self.output_enabled
+        self.output_enabled = not discard_output
+        try:
+            error = ClientAsync.aw(self.node.stop())
+            if error is not None:
+                raise Exception(f"Error {error['error_code']}")
+        finally:
+            self.output_enabled = output_enabled_orig
 
     @staticmethod
     def from_python_name(p_name):
@@ -257,7 +359,9 @@ class TDMConsole(code.InteractiveConsole):
                 ast.dump(node)
                 if isinstance(node.func, ast.Name):
                     fun_name = node.func.id
-                    fun_called.add(fun_name)
+                    if (fun_name not in ATranspiler.PREDEFINED_FUNCTIONS
+                        or fun_name in self.fun_defs):
+                        fun_called.add(fun_name)
                     if fun_name in self.fun_defs:
                         # call to a user-defined function
                         var_got |= self.fun_defs[fun_name]["in"]
