@@ -8,12 +8,14 @@
 import sys
 import os
 import getopt
+import re
 
 from tdmclient import ClientAsync
 from tdmclient.atranspiler import ATranspiler
+from tdmclient.module_thymio import ModuleThymio
 
 
-def help():
+def help(**kwargs):
     print("""Usage: python3 -m tdmclient.tools.run [options] [filename]
 Run program on robot, from file or stdin
 
@@ -21,14 +23,17 @@ Options:
   --debug n      display diagnostic information (0=none, 1=basic, 2=more, 3=verbose)
   --help         display this help message and exit
   --language=L   programming language (aseba or python); default=automatic
+  --nosleep      exit immediately (default with neither events nor print statement)
+  --nothymio     don't import the symbols of thymio library
   --robotid=I    robot id; default=any
   --robotname=N  robot name; default=any
   --scratchpad   also store program into the TDM scratchpad
+  --sleep        sleep forever (default with events or print statement)
   --sponly       store program into the TDM without running it
   --stop         stop program (no filename or stdin expected)
   --tdmaddr=H    tdm address (default: localhost or from zeroconf)
   --tdmport=P    tdm port (default: from zeroconf)
-""")
+""", **kwargs)
 
 
 if __name__ == "__main__":
@@ -41,24 +46,48 @@ if __name__ == "__main__":
     tdm_port = None
     robot_id = None
     robot_name = None
+    events = []
+    event_re = re.compile(r"^([^[]*)(\[([0-9]]*)\])?")
+    sleep = None  # True to sleep forever, False to exit immediately
+    import_thymio = True
+
+    print_statements = []
+    exit_received = None  # or exit status once received
+
+    def on_event_received(node, event_name, event_data):
+        if event_name == "_exit":
+            global exit_received
+            exit_received = event_data[0]
+        elif event_name == "_print":
+            print_id = event_data[0]
+            print_format, print_num_args = print_statements[print_id]
+            print_args = tuple(event_data[1 : 1 + print_num_args])
+            print_str = print_format % print_args
+            print(print_str)
+        else:
+            print(event_name + "".join(["," + str(d) for d in event_data]))
 
     try:
         arguments, values = getopt.getopt(sys.argv[1:],
                                           "",
                                           [
                                               "debug=",
+                                              "event=",
                                               "help",
                                               "language=",
+                                              "nosleep",
+                                              "nothymio",
                                               "robotid=",
                                               "robotname=",
                                               "scratchpad",
+                                              "sleep",
                                               "sponly",
                                               "stop",
                                               "tdmaddr=",
                                               "tdmport=",
                                           ])
     except getopt.error as err:
-        print(str(err))
+        print(str(err), file=sys.stderr)
         sys.exit(1)
     for arg, val in arguments:
         if arg == "--help":
@@ -66,14 +95,29 @@ if __name__ == "__main__":
             sys.exit(0)
         elif arg == "--debug":
             debug = int(val)
+        elif arg == "--event":
+            r = event_re.match(val)
+            if r is None:
+                help(file=sys.stderr)
+                sys.exit(1)
+            events.append((
+                r.group(1),
+                0 if r.group(3) is None else int(r.group(3)),
+            ))
         elif arg == "--language":
             language = val
+        elif arg == "--nosleep":
+            sleep = False
+        elif arg == "--nothymio":
+            import_thymio = False
         elif arg == "--robotid":
             robot_id = val
         elif arg == "--robotname":
             robot_name = val
         elif arg == "--scratchpad":
             scratchpad = 1
+        elif arg == "--sleep":
+            sleep = True
         elif arg == "--sponly":
             scratchpad = 2
         elif arg == "--stop":
@@ -85,7 +129,7 @@ if __name__ == "__main__":
 
     if stop:
         if len(values) > 0:
-            help()
+            help(file=sys.stderr)
             sys.exit(1)
     else:
         if len(values) == 0:
@@ -94,6 +138,9 @@ if __name__ == "__main__":
                 # try to transpile code from Python
                 try:
                     transpiler = ATranspiler()
+                    if import_thymio:
+                        transpiler.set_preamble("""from thymio import *
+""")
                     transpiler.set_source(program)
                     transpiler.transpile()
                     # successful, must be Python
@@ -108,14 +155,34 @@ if __name__ == "__main__":
                 # guess language from file extension
                 language = "python" if os.path.splitext(values[0])[1] == ".py" else "aseba"
         else:
-            help()
+            help(file=sys.stderr)
             sys.exit(1)
 
     status = 0
 
     if language == "python":
         # transpile from Python to Aseba
-        program = ATranspiler.simple_transpile(program)
+        modules = {
+            "thymio": ModuleThymio()
+        }
+        transpiler = ATranspiler()
+        transpiler.modules = {**transpiler.modules, **modules}
+        if import_thymio:
+            transpiler.set_preamble("""from thymio import *
+""")
+        transpiler.set_source(program)
+        transpiler.transpile()
+        program = transpiler.get_output()
+        print_statements = transpiler.print_format_strings
+        if len(print_statements) > 0:
+            events.append(("_print", 1 + transpiler.print_max_num_args))
+        if transpiler.has_exit_event:
+            events.append(("_exit", 1))
+        for event_name in transpiler.events:
+            events.append((event_name, transpiler.events[event_name]))
+
+    if sleep is None:
+        sleep = len(events) > 0
 
     with ClientAsync(tdm_addr=tdm_addr, tdm_port=tdm_port, debug=debug) as client:
 
@@ -129,11 +196,16 @@ if __name__ == "__main__":
                         status = 2
                 else:
                     if scratchpad < 2:
+                        if len(events) > 0:
+                            await node.register_events(events)
                         error = await node.compile(program)
                         if error is not None:
                             print(f"Compilation error: {error['error_msg']}")
                             status = 2
                         else:
+                            if len(events) > 0 and sleep:
+                                client.add_event_received_listener(on_event_received)
+                                await node.watch(events=True)
                             error = await node.run()
                             if error is not None:
                                 print(f"Run error {error['error_code']}")
@@ -143,6 +215,13 @@ if __name__ == "__main__":
                         if error is not None:
                             print(f"Scratchpad error {error['error_code']}")
                             status = 2
+                    if scratchpad < 2 and sleep:
+                        # expect events: wait forever or until _exit is received
+                        def wake():
+                            return exit_received is not None
+                        await client.sleep(-1, wake)
+                        await node.stop()
+                        status = exit_received
 
         client.run_async_program(prog)
 
